-- Name: EnhancedAnimalSystem
-- Author: Chissel

local modName = g_currentModName

EnhancedAnimalSystem = {}
EnhancedAnimalSystem.Settings = g_easUtils.loadXml()
EnhancedAnimalSystem.SellOvercrowding = true
EnhancedAnimalSystem.MaxHealthOvercrowdingDeduction = 5
EnhancedAnimalSystem.ModName = modName..".EnhancedAnimalSystem"

-- ################################# Lifecycle methods ####################################

function EnhancedAnimalSystem.prerequisitesPresent(specializations)
	g_easUtils:logText("prerequisitesPresent")
    return SpecializationUtil.hasSpecialization(PlaceableHusbandryAnimals, specializations);
end

function EnhancedAnimalSystem.registerEventListeners(placeableType)
	g_easUtils:logText("Register event listeners for", placeableType.name)
    SpecializationUtil.registerEventListener(placeableType, "onLoad", EnhancedAnimalSystem)
    SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", EnhancedAnimalSystem)
    SpecializationUtil.registerEventListener(placeableType, "onHusbandryAnimalsUpdate", EnhancedAnimalSystem)
    SpecializationUtil.registerEventListener(placeableType, "onPeriodChanged", EnhancedAnimalSystem)
    SpecializationUtil.registerEventListener(placeableType, "onHourChanged", EnhancedAnimalSystem)
end

function EnhancedAnimalSystem.registerOverwrittenFunctions(placeableType)
	g_easUtils:logText("Register overritten functions for", placeableType.name)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "getNumOfFreeAnimalSlots", EnhancedAnimalSystem.getNumOfFreeAnimalSlots)
end

-- ################################# Override functions ####################################

function EnhancedAnimalSystem:getNumOfFreeAnimalSlots()
    local spec = self.spec_husbandryAnimals
    local totalNumAnimals = self:getNumOfAnimals()
    if self.allowOvercrowding then
        local maxNumAnimals = spec.maxNumAnimals + math.floor(EnhancedAnimalSystem.Settings.MaxOvercrowdingPercentage * spec.maxNumAnimals)
        return math.max(maxNumAnimals - totalNumAnimals, 0)
    else
        return math.max(spec.maxNumAnimals - totalNumAnimals, 0)
    end
end

-- ################################# OnLoad ####################################

function EnhancedAnimalSystem:onLoad(savegame)
    if savegame ~= nil then
        g_easUtils:logText("onLoad for husbandry id: "..self.id)
        self.reproductionCalculated = true
        self.healthReducedForOvercrowding = true


        local key = savegame.key
        local path = key.."."..EnhancedAnimalSystem.ModName.."#allowOvercrowding"
        local allowOvercrowding = savegame.xmlFile:getValue(path, false)
        self.allowOvercrowding = allowOvercrowding
    end
end

-- ################################# OnHusbandryAnimalsUpdate ####################################

function EnhancedAnimalSystem:onHusbandryAnimalsUpdate(clusters)
	local spec = self.spec_husbandryAnimals
    local addedClusters = spec.clusterSystem.clustersToAdd

    EnhancedAnimalSystem:reduceHealthOnOvercrowdingIfNeeded(self)

    if self.reproductionCalculated == false then
        self.reproductionCalculated = true

        if addedClusters ~= nil and next(addedClusters) ~= nil then
            EnhancedAnimalSystem:removeAddedClusters(spec, addedClusters)
        end

        local motherClusters = EnhancedAnimalSystem.calculateMotherCluster(spec.clusterSystem.clusters, spec.animalTypeIndex)

        if motherClusters ~= nil then
            g_easUtils:logText("Start calculate reproduction clusters")
            EnhancedAnimalSystem:calculateReproduction(spec, motherClusters, spec:getNumOfFreeAnimalSlots(), self.isServer)
        end
    end
end

function EnhancedAnimalSystem:removeAddedClusters(spec, addedClusters)
    local numberOfRemovedClusters = 0

    for cluster, _ in pairs(addedClusters) do
        g_easUtils:removeItemFromTable(cluster, spec.clusterSystem.clusters)
        numberOfRemovedClusters = numberOfRemovedClusters + 1
    end

    g_easUtils:logText("Removed clusters: "..numberOfRemovedClusters)
end

function EnhancedAnimalSystem.calculateMotherCluster(clusters, animalTypeIndex)
    local motherCluster = {}
    local numberOfMotherClusters = 0

    for _, value in ipairs(clusters) do
        if value.reproduction == 0 and
            value.age > EnhancedAnimalSystem.reproductionMinAge(value.subTypeIndex) and
            value.health >= EnhancedAnimalSystem.reproductionMinHealth(value.subTypeIndex) and
            animalTypeIndex ~= AnimalSubType.CHICKEN_ROOSTER
        then
            table.insert(motherCluster, value)
            numberOfMotherClusters = numberOfMotherClusters + 1
        end
    end

    g_easUtils:logText("Mother clusters calculated: "..numberOfMotherClusters)

    return motherCluster
end

function EnhancedAnimalSystem.reproductionMinHealth(subTypeIndex)
    local subType = g_currentMission.animalSystem:getSubTypeByIndex(subTypeIndex)
    return subType.reproductionMinHealth * 100
end

function EnhancedAnimalSystem.reproductionMinAge(subTypeIndex)
    local subType = g_currentMission.animalSystem:getSubTypeByIndex(subTypeIndex)
    return subType.reproductionMinAgeMonth
end

function EnhancedAnimalSystem:calculateReproduction(spec, motherClusters, freeSpace, isServer)
    for _, cluster in ipairs(motherClusters) do
        local totalOffspring = EnhancedAnimalSystem.calculateTotalOffspring(cluster.numAnimals, spec.animalTypeIndex)

        if totalOffspring == 0 then
            g_easUtils:logText("No offspring")
        elseif freeSpace >= totalOffspring then
            g_easUtils:logText("Set cluster to totalOffspring "..totalOffspring)
            freeSpace = freeSpace - totalOffspring
            local newCluster = EnhancedAnimalSystem.createCluster(spec, totalOffspring, cluster.subTypeIndex)
            table.insert(spec.clusterSystem.clusters, newCluster)
        elseif freeSpace == 0 then
            g_easUtils:logText("Overcrowing of: "..totalOffspring)
            local newCluster = EnhancedAnimalSystem.createCluster(spec, 0, cluster.subTypeIndex)
            local sellPrice = newCluster:getSellPrice()
            EnhancedAnimalSystem.sellAnimalsIfAllowed(totalOffspring, spec.clusterHusbandry.placeable.ownerFarmId, sellPrice)
        else
            local overcrowding = totalOffspring - freeSpace
            g_easUtils:logText("Overcrowing of: "..overcrowding)

            local newCluster = EnhancedAnimalSystem.createCluster(spec, freeSpace, cluster.subTypeIndex)
            local sellPrice = newCluster:getSellPrice()
            local farmId = spec:getOwnerFarmId()
            table.insert(spec.clusterSystem.clusters, newCluster)

            EnhancedAnimalSystem.sellAnimalsIfAllowed(overcrowding, farmId, sellPrice, isServer)
            freeSpace = 0
        end
    end
end

function EnhancedAnimalSystem.calculateTotalOffspring(numAnimals, animalTypeIndex)
    local totalOffspring = 0

    for i=1, numAnimals do
        local multiplier = EnhancedAnimalSystem.reproductionMultiplierForAnimalType(animalTypeIndex)
        totalOffspring = totalOffspring + multiplier
    end

    return totalOffspring
end

function EnhancedAnimalSystem.sellAnimalsIfAllowed(numAnimals, farmId, sellPrice, isServer)
    if EnhancedAnimalSystem.SellOvercrowding then
        local pricePerAnimal = sellPrice * EnhancedAnimalSystem.Settings.SellingPricePercentage
        local amount = pricePerAnimal * numAnimals

        if isServer then
            g_currentMission:addMoneyChange(amount, farmId, MoneyType.SOLD_ANIMALS, true)

            g_easUtils:logText("Add "..amount.." money to farm with farmId: "..farmId)

            local farm = g_farmManager:getFarmById(farmId)
            farm:changeBalance(amount, MoneyType.SOLD_ANIMALS)

            g_easUtils:logText("Sold "..numAnimals.." for price of "..pricePerAnimal)
        else
            g_client:getServerConnection():sendEvent(EAS_AddMoneyEvent.new(amount, farmId))
            g_easUtils:logText("Send event for "..amount.." money for farm: "..farmId)
        end
    end
end

function EnhancedAnimalSystem.priceForAnimalType(animalTypeIndex)
    if animalTypeIndex == AnimalType.PIG then
        return EnhancedAnimalSystem.Settings.PigletsSoldPrice
    elseif animalTypeIndex == AnimalType.COW then
        return EnhancedAnimalSystem.Settings.CalfSoldPrice
    elseif animalTypeIndex == AnimalType.SHEEP then
        return EnhancedAnimalSystem.Settings.LambSoldPrice
    elseif animalTypeIndex == AnimalType.HORSE then
        return EnhancedAnimalSystem.Settings.FoalSoldPrice
    elseif animalTypeIndex == AnimalType.CHICKEN then
        return EnhancedAnimalSystem.Settings.ChickenSoldPrice
    end

    return 1
end

function EnhancedAnimalSystem.createCluster(spec, numAnimals, subTypeIndex)
    local cluster = g_currentMission.animalSystem:createClusterFromSubTypeIndex(subTypeIndex)
    cluster.numAnimals = numAnimals
    cluster.age = 0
    return cluster
end

function EnhancedAnimalSystem.reproductionMultiplierForAnimalType(animalTypeIndex)
    if animalTypeIndex == AnimalType.PIG then
        return EnhancedAnimalSystem.Settings.PigletProbabilities[math.random(#EnhancedAnimalSystem.Settings.PigletProbabilities)]
    elseif animalTypeIndex == AnimalType.COW then
        return EnhancedAnimalSystem.Settings.CalfProbabilities[math.random(#EnhancedAnimalSystem.Settings.CalfProbabilities)]
    elseif animalTypeIndex == AnimalType.SHEEP then
        return EnhancedAnimalSystem.Settings.LambProbabilities[math.random(#EnhancedAnimalSystem.Settings.LambProbabilities)]
    elseif animalTypeIndex == AnimalType.HORSE then
        return EnhancedAnimalSystem.Settings.FoalProbabilities[math.random(#EnhancedAnimalSystem.Settings.FoalProbabilities)]
    elseif animalTypeIndex == AnimalType.CHICKEN then
        return EnhancedAnimalSystem.Settings.ChickenProbabilities[math.random(#EnhancedAnimalSystem.Settings.ChickenProbabilities)]
    end

    return 1
end

function EnhancedAnimalSystem.hasMotherCluster(cluster, clusters)
    local numOfMothers = 0

    for _, compareCluster in ipairs(clusters) do
        if compareCluster:getAge() > 5 and compareCluster.reproduction == 0 and compareCluster.subTypeIndex == cluster.subTypeIndex then
            g_easUtils:logText("Mother cluster detected")
            numOfMothers = numOfMothers + compareCluster.numAnimals
        end
    end

    return cluster.numAnimals <= numOfMothers
end

-- ################################# OnFinalizePlacement ####################################

function EnhancedAnimalSystem:onFinalizePlacement()
    g_easUtils:logText("onFinalizePlacement for husbandry id: "..self.id)
    self.reproductionCalculated = true
    self.healthReducedForOvercrowding = true
end

-- ################################# OnPeriodChanged ####################################

function EnhancedAnimalSystem:onPeriodChanged()
    self.reproductionCalculated = false
    g_easUtils:logText("ReproductionCalculatedForHusbandryIds resetted")

    EnhancedAnimalSystem:calculateOldAnimalsDies(self)
end

-- ################################# Overcrowding ####################################

function EnhancedAnimalSystem:onHourChanged(currentHour)
    self.healthReducedForOvercrowding = false
    g_easUtils:logText("onHourChanged")
end

function EnhancedAnimalSystem:reduceHealthOnOvercrowdingIfNeeded(spec)
    if spec.healthReducedForOvercrowding then
        return
    end
    spec.healthReducedForOvercrowding = true

	local husbandrySpec = spec.spec_husbandryAnimals
    local totalNumAnimals = husbandrySpec:getNumOfAnimals()
    local clusters = husbandrySpec.clusterSystem.clusters
    if husbandrySpec.maxNumAnimals < totalNumAnimals then
        local overcrowdingInPercentage = (totalNumAnimals - husbandrySpec.maxNumAnimals) / 100
        local healthDeduction = math.floor(EnhancedAnimalSystem.MaxHealthOvercrowdingDeduction * overcrowdingInPercentage) + 10

        for _, cluster in ipairs(clusters) do
            local healthValue = cluster.health
            if healthValue >= healthDeduction then
                cluster.health = healthValue - healthDeduction
            else
                cluster.health = 0
            end
        end
    end
end

-- ################################# Old animals die ####################################

function EnhancedAnimalSystem:calculateOldAnimalsDies(spec)
    local husbandrySpec = spec.spec_husbandryAnimals
    local clusters = husbandrySpec.clusterSystem.clusters

    for _, cluster in ipairs(clusters) do
        if cluster.age > EnhancedAnimalSystem.maxAgeForAnimal(husbandrySpec.animalTypeIndex) then
            g_easUtils:logText("Cluster is too old. Maybe some animals will die.")

            local numOfDeadAnimals = EnhancedAnimalSystem.numOfDeadAnimals(cluster.numAnimals)
            cluster.numAnimals = cluster.numAnimals - numOfDeadAnimals
        end
    end
end

function EnhancedAnimalSystem.numOfDeadAnimals(numAnimals)
    local deadAnimals = 0

    for i=1, numAnimals do
        local randomNumber = math.random(100)
        if randomNumber <= EnhancedAnimalSystem.Settings.AnimalDiedPercentage then
            deadAnimals = deadAnimals + 1
        end
    end

    g_easUtils:logText(deadAnimals.." animals have died due to their age")

    return deadAnimals
end

function EnhancedAnimalSystem.maxAgeForAnimal(animalTypeIndex)
    if animalTypeIndex == AnimalType.PIG then
        return EnhancedAnimalSystem.Settings.PigMaxAge
    elseif animalTypeIndex == AnimalType.COW then
        return EnhancedAnimalSystem.Settings.CowMaxAge
    elseif animalTypeIndex == AnimalType.HORSE then
        return EnhancedAnimalSystem.Settings.HorseMaxAge
    elseif animalTypeIndex == AnimalType.SHEEP then
        return EnhancedAnimalSystem.Settings.SheepMaxAge
    elseif animalTypeIndex == AnimalType.CHICKEN then
        return EnhancedAnimalSystem.Settings.ChickenMaxAge
    end

    return 120
end

-- ################################# Save ####################################

function EnhancedAnimalSystem.registerSavegameXMLPaths(schema, basePath)
    local path = basePath.."#allowOvercrowding"
    schema:register(XMLValueType.BOOL, path, "Allow overcrowding for husbandry", false)
    g_easUtils:logText("registerSavegameXMLPaths")
end
-- FS22_EnhancedAnimalSystem_beta.EnhancedAnimalSystem
function EnhancedAnimalSystem:saveToXMLFile(xmlFile, key, usedModNames)
    local path = key.."#allowOvercrowding"
    xmlFile:setValue(path, self.allowOvercrowding)
    g_easUtils:logText("saveToXMLFile")
end
